import { IAccessory } from "./models/iAccessory";
import { ISequence } from "./models/iConfig";
import Api = require("node-hue-api/lib/api/Api");
import Light = require("node-hue-api/lib/model/Light");
import LightState = require("node-hue-api/lib/model/lightstate/LightState");
import { Sleep } from "./sleep";

let Service: HAPNodeJS.Service;
let Characteristic: HAPNodeJS.Characteristic;

export interface IChaseProps {
    name: string;
    api: any;
    log: any;
    homebridge: any;
    sequence: ISequence;
    hue: Api
}

export class Chase implements IAccessory {
    private _api: any;
    private _homebridge: any;
    private _log: any = {};

    //Service fields
    private _switchService: HAPNodeJS.Service;
    private _infoService: HAPNodeJS.Service;

    private _sequence: ISequence;

    private _isActive: boolean;
    private _hue: Api;

    private _lights: Array<Light> = [];

    constructor(props: IChaseProps) {
        //Assign class variables
        this._log = props.log;
        this._api = props.api;
        Service = props.api.hap.Service;
        Characteristic = props.api.hap.Characteristic;
        this.name = props.name;
        this._homebridge = props.homebridge;

        this._sequence = props.sequence;

        this._isActive = false;
        this._hue = props.hue;

        this.platformAccessory = new this._homebridge.platformAccessory(this.name, this.generateUUID(), this._homebridge.hap.Accessory.Categories.SWITCH);

        //@ts-ignore
        this._infoService = new Service.AccessoryInformation();
        this._infoService.setCharacteristic(Characteristic.Manufacturer, "The Watson Project")
        this._infoService.setCharacteristic(Characteristic.Model, "Hue Chase")
        this._infoService.setCharacteristic(Characteristic.SerialNumber, "123-456-789");

        this._switchService = new Service.Switch(
            this.name,
            'switchService'
        )

        this._switchService.getCharacteristic(Characteristic.On)
            //@ts-ignore
            .on("set", this.onSwitchSet)
            .on("get", this.onSwitchGet);

        this.getLights();

    }

    /**
     * Required by homebridge.
     */
    public name: string;

    public platformAccessory: any;

    /**
     * Called by homebridge to gather services.
     */
    public getServices = (): Array<HAPNodeJS.Service> => {
        return [this._switchService!];
    }

    /**
     * Handler for switch set event
     * @param callback The callback function to call when complete
     */
    private onSwitchSet = (activeState: boolean, callback: (error?: Error | null | undefined) => void) => {
        activeState ? this.chase() : this.stop();
        return callback();
    }

    /**
     * Handler for switch get event
     * @param callback The callback function to call when complete
     */
    private onSwitchGet = (callback: (error: Error | null, value: boolean) => void) => {
        return callback(null, this._isActive);
    }

    /**
     * Popuplates internal lights array using the configuration values
     */
    private getLights = async () => {
        //Get lights
        const lightPromises: Array<Promise<void>> = this._sequence.lights.map(async (value: string) => {
            //@ts-ignore
            const light: Light = await this._hue.lights.getLightByName(value)
            this._lights.push(light);
        });
        await Promise.all(lightPromises);
    }

    /**
     * Execute chase sequence
     */
    private chase = async () => {
        if (this._lights.length === 0) {
            await this.getLights();
        }
        const lightState = new LightState();
        let idx = 0;
        this._isActive = true;
        while (this._isActive) {
            const rgb = this.hexToRgb(this._sequence.colors[idx])
            lightState
                .on(true)
                .brightness(this._sequence.intensity)
                .transitionInMillis(this._sequence.transitionTime)
                .rgb(rgb?.red, rgb?.green, rgb?.blue);

            try {
                await Promise.all(
                    this._lights.map(async (value: Light) => {
                        await this._hue.lights.setLightState(value.id, lightState);
                    })
                );
            } catch (err) {
                this._log(`Error while changing lights: ${err}`)
            }


            await Sleep(this._sequence.transitionTime);
            if (idx == this._sequence.colors.length) {
                idx = 0;
            } else {
                idx++;
            }
        }
    }

    /**
     * Stop chase sequence
     */
    private stop = () => {
        this._isActive = false;
    }

    /**
     * Helper function to convert a hex string to rgb
     * @param hex hex string starting with "#"
     */
    private hexToRgb = (hex: string): { red: number, green: number, blue: number } | null => {
        var result = /^#?([a-f\d]{2})([a-f\d]{2})([a-f\d]{2})$/i.exec(hex);
        return result ? {
            red: parseInt(result[1], 16),
            green: parseInt(result[2], 16),
            blue: parseInt(result[3], 16)
        } : null;
    }

    /**
     * Helper function to generate a UUID
     */
    private generateUUID(): string { // Public Domain/MIT
        var d = new Date().getTime();
        if (typeof performance !== 'undefined' && typeof performance.now === 'function') {
            d += performance.now(); //use high-precision timer if available
        }
        return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function (c) {
            var r = (d + Math.random() * 16) % 16 | 0;
            d = Math.floor(d / 16);
            return (c === 'x' ? r : (r & 0x3 | 0x8)).toString(16);
        });
    }
}